"""

.. todo::
    # get cas number from chemical name or formula
    # get molecular weight of chemical from chemical name or formula

"""


import os
ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
